// Copyright (C) 2018 Avatao.com Innovative Learning Kft.
// All Rights Reserved. See LICENSE file for details.

import { ChangeDetectorRef, Component, OnInit, EventEmitter, Output } from '@angular/core';
import { MessageConfig, MessageData, MessagesMessage, MessageQueueMessage } from '../message-types/bot-messages';
import { MarkdownService } from '../services/markdown.service';
import { WebSocketService } from '../services/websocket.service';

import { config } from '../config';
import { Subject, Observer, BehaviorSubject } from 'rxjs';

@Component({
  selector: 'app-messages',
  templateUrl: './messages.component.html',
  styleUrls: ['./messages.component.scss']
})
export class MessagesComponent implements OnInit {
  @Output() newMessageEvent: EventEmitter<any> = new EventEmitter();
  newMessage: Subject<MessageData> = new Subject<MessageData>();
  messageInQueue = true;

  originator: string = 'avataobot';
  messages: MessageData[] = [];
  messageQueueAttender: MessageQueueAttender;

  constructor(
    private markdownService: MarkdownService,
    private websocketService: WebSocketService,
    private changeDetectorRef: ChangeDetectorRef
  ) {
    this.messageQueueAttender = new MessageQueueAttender(this.newMessage);
  }

  ngOnInit() {
    this.newMessage.subscribe(
      (message) => {
        this.writeMessage(message);
        this.newMessageEvent.emit();
    });
    this.messageQueueAttender.messageInQueue.subscribe(
      (value) => this.messageInQueue = value
    );

    this.websocketService.connect();
    this.websocketService.observeKey<MessagesMessage>('message.send').subscribe(
      (event) => this.newMessage.next(event)
    );
    this.websocketService.observeKey<MessageQueueMessage>('message.queue').subscribe(
      (event) => this.handleQueueMessage(event)
    );
    this.websocketService.observeKey<MessageConfig>('message.config').subscribe(
      (event) => this.configureMessages(event)
    );
  }

  writeMessage(data: MessageData) {
    this.transformMessage(data);
    this.messages.push(data);
    this.changeDetectorRef.detectChanges();
  }

  transformMessage(data: MessageData) {
    data.message = this.convertMarkdownToHTML(data.message);
    if (!data.originator) {
      data.originator = this.originator;
    }
    if (!data.timestamp) {
      data.timestamp = new Date();
    }
  }

  handleQueueMessage(data: MessageQueueMessage) {
    this.messageQueueAttender.queueMessages(data.value);
  }

  configureMessages(data: MessageConfig) {
    if (data.originator) {
      this.originator = data.originator;
    }
  }

  convertMarkdownToHTML(text: string) {
    return this.markdownService.convertToHtml(text);
  }
}


class MessageQueueAttender {
  public messageInQueue = new BehaviorSubject<boolean>(false);

  private readonly charPerSecond: number;
  private lastMessageLength = 0;
  private queue: MessageData[] = [];

  constructor(private messageEmitter: Observer<MessageData>, wordPerMinute: number = config.messages.messageQueueWPM) {
    const charPerMinute = wordPerMinute * 5;
    this.charPerSecond = charPerMinute / 60;
  }

  queueMessages(messages: Array<MessageData>) {
    this.queue = this.queue.concat(messages);
    this.attendQueue();
  }

  private processMessage() {
    if (this.queue.length > 0) {
      const lastMessage = this.queue.shift();
      this.lastMessageLength = lastMessage.message.length;
      this.messageEmitter.next(lastMessage);
    }
    if (this.queue.length === 0) {
      this.lastMessageLength = 0;
      this.messageInQueue.next(false);
    }
  }

  private attendQueue() {
    if (this.queue.length > 0) {
      this.messageInQueue.next(true);
      const timeoutSeconds = this.lastMessageLength / this.charPerSecond;
      setTimeout(
        () => {
          this.processMessage();
          this.attendQueue();
        },
        timeoutSeconds * 1000
      );
    }
  }
}
